/*----------------------------------------------------------------------------*
 *                                                                            *
 *   HPF Array Combining Scatter Function                                     *
 *                                                                            *
 *     - Author: Seungjo Bae & Khaled A. Alsabti                              *
 *                                                                            *
 *----------------------------------------------------------------------------*/

/******************************************************************************

  History

     1) Check / * operator * / when take care of operator (8/9/97)

     2) Be careful when use malloc in a function (8/9/97)

     3) IPARITY can not be supported, because no identity value (8/10/97)
               
     4) INT_comb_1 : Use #define instead of function (8/10/97)
          - Check INT_comb_1(.....)

     5) Optimize operation by vectorization (8/11/97)
          - Check switch (OP_TYPE) { / * vector operator * / 
                     .
                     .
		  }
          - Check INT_vector_comb_1

     -----

     6) Interface with Bryan (8/27/97)

 ******************************************************************************/
 
#include "raw_global.h"

float Temp_Real;			/* temporary value for PRODUCT */

/******************************************************************************
 *                                                                            *
 *  Inline Function : COMPLEX_comb_1(l, r) 
 *                                                                            *
 ******************************************************************************/

/*--- cause error in PLC (mesg decomp for sparse bucket, but no idea why ---*/
#define COMPLEX_PRODUCT(l,r) \
  Temp_Real = (l).Real * (r).Real - (l).Imag * (r).Imag; \
  (l).Imag  = (l).Real * (r).Imag + (l).Imag * (r).Real; \
  (l).Real  = Temp_Real;

#define COMPLEX_SUM(l,r) \
  (l).Imag += (r).Imag;  \
  (l).Real += (r).Real; 

#define COMPLEX_comb_1(l, r)          		       \
  switch (OP_TYPE) {             		       \
                                 		       \
  case COPY:                     		       \
    (l).Real = (r).Real;                   	       \
    (l).Imag = (r).Imag;                   	       \
    break;                       		       \
                                 		       \
  case PRODUCT:                  		       \
    Temp_Real = (l).Real * (r).Real - (l).Imag * (r).Imag;  \
    (l).Imag  = (l).Real * (r).Imag + (l).Imag * (r).Real;  \
    (l).Real  = Temp_Real; \
    break;                       		       \
                                 		       \
  case SUM:                      		       \
    (l).Real += (r).Real;                  	       \
    (l).Imag += (r).Imag;                  	       \
    break;                       		       \
  }


/******************************************************************************
 *                                                                            *
 *  Function : COMPLEX_vector_comb_1(l, r, size) 
 *                                                                            *
 ******************************************************************************/

void COMPLEX_vector_comb_1(
   COMPLEX_Data_Type *l,
   COMPLEX_Data_Type *r,
   int size)
{
  int j;

  switch (OP_TYPE) {    
                        
  case COPY:            
    for (j=0; j < size ; j++) {
      l->Real = r->Real;
      l->Imag = r->Imag;
      ++l; ++r;
    }
    break;              
                        
  case PRODUCT:         
    for (j=0; j < size ; j++) {
      Temp_Real = l->Real * r->Real - l->Imag * r->Imag; 
      l->Imag   = l->Real * r->Imag + l->Imag * r->Real; 
      l->Real   = Temp_Real;
      ++l; ++r;
    }
    break;              
                        
  case SUM:             
    for (j=0; j < size ; j++) {
      l->Real += r->Real;
      l->Imag += r->Imag;
      ++l; ++r;
    }
    break;              
  }
}


/******************************************************************************
 *                                                                            *
 *  Function :  COMPLEX_nplc_tsc
 *                                                                            *
 ******************************************************************************/

void COMPLEX_nplc_tsc(
  COMPLEX_Data_Type *source_array,         
  int source_lidx[],
  int NP,                     /* s_L_all */                                 

  int index_lidx[],
  int index_pnum[],   
 
  COMPLEX_Data_Type *target_array,     
  int MP,                     /* t_L_all */                                 

  int algorithm)
{

  COMPLEX_Buffer_Type *recvbuf;	/* Alternatively, we can declare it as */
  COMPLEX_Data_Type *sendbuf2;	/* global variables for each type */

  void               COMPLEX_scan();
  void               stage1_pre();
  COMPLEX_Buffer_Type*   COMPLEX_stage1_comm();
  void               stage2_pre();
  COMPLEX_Data_Type*     COMPLEX_stage2_local_hash();
  void               stage2_pre_segment();
  void               COMPLEX_stage2_comm();

  /*--------------------------- Define data unit ----------------------------*/

  Data_Unit = sizeof(COMPLEX_Data_Type);
  Buffer_Unit = sizeof(COMPLEX_Buffer_Type);

  /*--------------------------- Define identity ----------------------------*/

  switch (OP_TYPE) {      
                          
  case COPY:              
    COMPLEX_identity.Real = 0;
    COMPLEX_identity.Imag = 0;
    break;                
                          
  case PRODUCT:           
    COMPLEX_identity.Real = 1;
    COMPLEX_identity.Imag = 0;
    break;                

  case SUM:               
    COMPLEX_identity.Real = 0;
    COMPLEX_identity.Imag = 0;
    break;                
                          
  default:                                
    printf("\n Improper integer operator = %d !\n", OP_TYPE);
    exit(0);                                  
  }

  /*------------------------------ Initial Scan -----------------------------*/

  COMPLEX_scan(source_array, source_lidx, NP,
	   index_lidx, index_pnum, 
	   target_array, MP,
	   algorithm);                                       
/***
print_array1_complex(TARGET, target_array, "NLC-TSC",  'c');
***/

  /*-------------------------------- Stage I -------------------------------*/

  stage1_pre();              

  recvbuf =
    COMPLEX_stage1_comm(source_array, source_lidx, NP,
		    index_lidx, index_pnum,                          
		    target_array, MP,
		    algorithm);

  /*-------------------------------- Stage I -------------------------------*/

  stage2_pre();	
		
  sendbuf2 = 
    COMPLEX_stage2_local_hash(source_array, source_lidx, NP,
			  index_lidx, index_pnum,                          
			  target_array, MP,
			  algorithm, recvbuf); 

  stage2_pre_segment();	

  COMPLEX_stage2_comm(source_array, source_lidx, NP,
		  index_lidx, index_pnum,                          
		  target_array, MP,
		  algorithm, sendbuf2);

  if (recvbuf != NULL)                    free(recvbuf);
  if (sendbuf2 != NULL)                   free(sendbuf2);

}


/******************************************************************************
 *                                                                            *
 *  Function : COMPLEX_scan
 *                                                                            *
 ******************************************************************************/

void COMPLEX_scan(
  COMPLEX_Data_Type *source_array,         
  int source_lidx[],
  int NP,                     /* s_L_all */                                 

  int index_lidx[],
  int index_pnum[],   
 
  COMPLEX_Data_Type *target_array,     
  int MP,                     /* t_L_all */                                 

  int algorithm)
{

  int pn, lidx, i, B;

  /*---------- Counting the number of threads writing to each bucket ---------*/

  for (i=0 ; i < NP ; i++)  {

    pn = index_pnum[i];
    lidx = index_lidx[i];

    if ( pn == my_pnum ) {
      /*target_array[ lidx ] += source_array[ source_lidx[i] ];*/ /* operator */
      COMPLEX_comb_1(target_array[ lidx ], 
		     source_array[ source_lidx[i] ]); 
      cache_B[i] = -1;
    }
    else {
      B = pn*K + (lidx/MP_div_K);
      cache_B[i]    = B;
      count_bucket[ B ]++;
    }
  }

}


/******************************************************************************
 *                                                                            *
 *  Function : COMPLEX_stage1_comm
 *                                                                            *
 ******************************************************************************/

COMPLEX_Buffer_Type* COMPLEX_stage1_comm(
  COMPLEX_Data_Type *source_array,         
  int source_lidx[],
  int NP,                     /* s_L_all */                                 
			                                                    
  int index_lidx[],	                                                    
  int index_pnum[],   	                                                    
 			                                                    
  COMPLEX_Data_Type *target_array,
  int MP,           	      /* t_L_all */                                 

  int algorithm)
{

  int ind, ind1;
  int B, i, j;

  COMPLEX_Buffer_Type *sendbuf, *recvbuf;    
  COMPLEX_Buffer_Type *recvbufSP,*sendbufSP;

  /*-------------------- we send value + index -----------------------------*/

  /*=== Format: (value, pointer) ===*/

  sendbuf = (COMPLEX_Buffer_Type *) malloc(total_send*sizeof(COMPLEX_Buffer_Type));
  recvbuf = (COMPLEX_Buffer_Type *) malloc(total_recv*sizeof(COMPLEX_Buffer_Type));

  sendbufSP = (COMPLEX_Buffer_Type *) malloc(total_sendSP*sizeof(COMPLEX_Buffer_Type));
  recvbufSP = (COMPLEX_Buffer_Type *) malloc(total_recvSP*sizeof(COMPLEX_Buffer_Type));

  if ( ( total_send   > 0 && sendbuf == NULL   ) || 
       ( total_recv   > 0 && recvbuf == NULL   ) ||
       ( total_sendSP > 0 && sendbufSP == NULL ) ||
       ( total_recvSP > 0 && recvbufSP == NULL ) ) {
    printf(" Out of Memory: recvbufSP in nlc_raw_stage1()\n"); 
    exit(0);
  }		

  /*=========================================================================*
   *===                     B E G I N   O F   P L C                       ===*   
   *=========================================================================*/

if (PLC) {

  /*----------------------- Initialize hash table  -------------------------*/

  for (i=0; i < KPS ; i++)
    if ( LSparse[i] == 0 ) {
  
      ind = bucket_Start[i];
      ind1 = i*MP_div_K; 
  
      if ( Sparse[i] == 0 )
	for (j=ind; j < ind+MP_div_K ; ++j) {
	  sendbuf[j].data = COMPLEX_identity;
	  sendbuf[j].ptr  = ind1++;            /* global index in BLOCK */
                                               /* prank in Block-Cyclic */
	}
      else {
	ind1 = ind1%MP;
	for (j=ind; j < ind+MP_div_K ; ++j) {
	  sendbufSP[j].data = COMPLEX_identity;
	  sendbufSP[j].ptr = ind1++;           /* local index */
	}
      }
    }

  /*-------------------- Building the send buffer  ---------------------------*/


  for (i=0 ; i < NP ; i++)  {
  
    B = cache_B[i];
  
    if ( B != -1 ) {
  
  	if ( Sparse[B] == 0)  {   /* globally dense */
  	  if ( LSparse[B] == 1 ) {   /* locally sparse */
  
  	    ind = bucket_Start[B]++;
  
  	    sendbuf[ind].data = source_array[ source_lidx[i] ];
            /*--- GENERAL ---*/
  	    sendbuf[ind].ptr  = index_pnum[i]*virtual_MP + index_lidx[i]; 
                                                            /* prank */
  	  }
  	  else  {      /* locally dense */
  	    ind = bucket_Start[B] + index_lidx[i]%MP_div_K;
  	    /*sendbuf[ind].data += source_array[ source_lidx[i] ];*/
	    /* operator */
	    COMPLEX_comb_1(sendbuf[ind].data, source_array[ source_lidx[i] ]);
  	  }
  	} /* globally dense */ 
  
  	else {/* globally sparse */
  	  if ( LSparse[B] == 1 ) {  /* locally sparse bucket */
  
  	    ind =bucket_Start[B]++;
  
  	    sendbufSP[ind].data = source_array[ source_lidx[i] ];
  	    sendbufSP[ind].ptr  = index_lidx[i];
  	  }
  	  else {  /* locally dense */
  	    ind = bucket_Start[B] + index_lidx[i]%MP_div_K; 
  	    /*sendbufSP[ind].data += source_array[ source_lidx[i] ];*/
	    /* operator */
	    COMPLEX_comb_1(sendbufSP[ind].data, source_array[ source_lidx[i] ]);
  	  }
  	} /* globally sparse */
  
    }
  
  } /* for i */

}
  /*=========================================================================*
   *===                        E N D   O F   P L C                        ===*   
   *=========================================================================*/


  /*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*
   *=-=                  B E G I N   O F   N L C                          =-=*
   *=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

else {

  /*-------------------- Building the send buffer  ---------------------------*/

  for (i=0 ; i < NP ; i++)  {
  
    B = cache_B[i];
  
    if ( B != -1 ) {
  
  	if ( Sparse[B] == 0 )  {/* dense bucket */
  	  sendbuf[ bucket_Start[B] ].data = source_array[ source_lidx[i] ];
  
        /*- grank version:                                   -*/
  	  /*- sendbuf[ bucket_Start[B] * 2 + 1 ] = index_array[i]; -*/
  
  	  /*---------- prank instead of grank (1/22/97) --------*/
          /*--- GENERAL ---*/
  	  sendbuf[ bucket_Start[B] ].ptr = index_pnum[i]*virtual_MP
                                         + index_lidx[i];
  	  ++bucket_Start[B];
  	}
  	else {
  	  sendbufSP[ bucket_Start[B] ].data = source_array[ source_lidx[i] ];
  	  sendbufSP[ bucket_Start[B] ].ptr  = index_lidx[i];
  	  ++bucket_Start[B];
  	}
  
    }
  } /* for i */

}

  /*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*
   *=-=                    E N D   O F   N L C                            =-=*
   *=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

  /*----------------------   For Sparse Buckets  -----------------------------*/

  M_to_M_PC(sendbufSP,sendcountSP,sdisplsSP,sizeof(COMPLEX_Buffer_Type),
	    recvbufSP,recvcountSP,rdisplsSP,algorithm);

  /*---------------------   First stage communcations   ----------------------*/

  M_to_M_PC(sendbuf,sendcount,sdispls,sizeof(COMPLEX_Buffer_Type),
	    recvbuf,recvcount,rdispls,algorithm);

  /*----------------------   For Sparse Buckets  -----------------------------*/

  switch (OP_TYPE) {		/* vector operator */
                        
  case COPY:            
    for (i=0 ; i < total_recvSP ; i++) 
      target_array[recvbufSP[i].ptr] = recvbufSP[i].data;
    break;              
                        
  case PRODUCT:         
    for (i=0 ; i < total_recvSP ; i++) {
      /*** cause error in PLC. BUT no idea why!
      COMPLEX_PRODUCT(target_array[recvbufSP[i].ptr], recvbufSP[i].data);
      ***/
      Temp_Real =
	target_array[recvbufSP[i].ptr].Real * recvbufSP[i].data.Real 
	  - target_array[recvbufSP[i].ptr].Imag * recvbufSP[i].data.Imag;  
      target_array[recvbufSP[i].ptr].Imag = 
	target_array[recvbufSP[i].ptr].Real * recvbufSP[i].data.Imag 
	  + target_array[recvbufSP[i].ptr].Imag * recvbufSP[i].data.Real;  
      target_array[recvbufSP[i].ptr].Real = Temp_Real;
    }
    break;              
                        
  case SUM:             
    for (i=0 ; i < total_recvSP ; i++) {
      /***/
      COMPLEX_SUM(target_array[recvbufSP[i].ptr], recvbufSP[i].data);
      /***/
      /***
      target_array[recvbufSP[i].ptr].Real += recvbufSP[i].data.Real;
      target_array[recvbufSP[i].ptr].Imag += recvbufSP[i].data.Imag;
      ***/
    }

    break;              
  } /* switch */

   /*for (i=0 ; i < total_recvSP ; i++)*/
     /*target_array[recvbufSP[i].ptr] += recvbufSP[i].data;*/   /* operator */
     /*COMPLEX_comb_1(target_array[recvbufSP[i].ptr], recvbufSP[i].data);*/

  /*=========================================================================*
   *===                     B E G I N   O F   P L C                       ===*   
   *=========================================================================*/

if (PLC)
  free(LSparse);

  /*=========================================================================*
   *===                        E N D   O F   P L C                        ===*   
   *=========================================================================*/

  if (sendbuf != NULL)                    free(sendbuf);
  if (sendbufSP != NULL)                  free(sendbufSP);
  if (recvbufSP != NULL)                  free(recvbufSP);
                                    
  return recvbuf;

}


/******************************************************************************
 *                                                                            *
 *  Function : COMPLEX_stage2_local_hash
 *                                                                            *
 ******************************************************************************/

COMPLEX_Data_Type* COMPLEX_stage2_local_hash(
  COMPLEX_Data_Type *source_array,         
  int source_lidx[],
  int NP,                     /* s_L_all */                                 
			                                                    
  int index_lidx[],	                                                    
  int index_pnum[],   	                                                    
 			                                                    
  COMPLEX_Data_Type *target_array, 
  int MP,        	      /* t_L_all */                                 

  int algorithm,

  COMPLEX_Buffer_Type *recvbuf)
{
  int ind, B, i, j;

  COMPLEX_Data_Type *sendbuf2;

  sendbuf2 = (COMPLEX_Data_Type *) malloc(Size*sizeof(COMPLEX_Data_Type));
  if ( Size > 0 && sendbuf2 == NULL ) {
    printf(" Out of Memory: sendbuf2 in raw_stage2()\n"); 
    exit(0);
  }		

  /*--- Initialize sendbuf2 buckets for full bucket and last_bucket if it is
        partial ---*/

  for (i=0; i < Size ; i++) 
    sendbuf2[i] = COMPLEX_identity;   /*----- It should be IDENTITY value -----*/

  /*------ copy the values to the sending buffer and resolve collision -----*/

  switch (OP_TYPE) { /* vector operator */
                    
  case COPY:            
    for (i=0 ; i < total_recv ; i++)  { 
      B = recvbuf[i].ptr/MP_div_K;   ind = recvbuf[i].ptr % MP_div_K;
      sendbuf2[ptr_bucket[B]+ind] = recvbuf[i].data;
    }
    break;              
                    
  case PRODUCT:         
    for (i=0 ; i < total_recv ; i++)  { 
      B = recvbuf[i].ptr/MP_div_K;   ind = recvbuf[i].ptr % MP_div_K;
      COMPLEX_PRODUCT(sendbuf2[ptr_bucket[B]+ind], recvbuf[i].data);
    }
    break;              
                    
  case SUM:             
    for (i=0 ; i < total_recv ; i++)  { 
      B = recvbuf[i].ptr/MP_div_K;   ind = recvbuf[i].ptr % MP_div_K;
      /***/
      COMPLEX_SUM(sendbuf2[ptr_bucket[B]+ind], recvbuf[i].data);
      /***/
      /***
      sendbuf2[ptr_bucket[B]+ind].Real += recvbuf[i].data.Real;
      sendbuf2[ptr_bucket[B]+ind].Imag += recvbuf[i].data.Imag;
      ***/
    }
    break;              

  } /* switch */

  /***for (i=0 ; i < total_recv ; i++)  {  ***/
    /*B = recvbuf[i].ptr >> shift_amount; <<=== power of 2 */
    /***B = recvbuf[i].ptr/MP_div_K;***/
    /*ind = recvbuf[i].ptr & Mask_MP_div_K; <<=== power of 2 */
    /***ind = recvbuf[i].ptr % MP_div_K;***/
    /*** sendbuf2[ptr_bucket[B]+ind] += recvbuf[i].data; ***/    /* operator */
    /*** COMPLEX_comb_1(sendbuf2[ptr_bucket[B]+ind], recvbuf[i].data); ***/
  /*** }  ***/

  return sendbuf2;

}



/******************************************************************************
 *                                                                            *
 *  Function : COMPLEX_stage2_comm
 *                                                                            *
 ******************************************************************************/

void COMPLEX_stage2_comm(
  COMPLEX_Data_Type *source_array,         
  int source_lidx[],
  int NP,                     /* s_L_all */                                 
			                                                    
  int index_lidx[],	                                                    
  int index_pnum[],   	                                                    
 			                                                    
  COMPLEX_Data_Type *target_array, 
  int MP,        	      /* t_L_all */                                 

  int algorithm,
  COMPLEX_Data_Type *sendbuf2)
{
  int ind, end, destPN, i, j;
  int first, last_plus1;

  COMPLEX_Data_Type *temp;
  COMPLEX_Data_Type *recvbuf2;    

  void COMPLEX_Segmented_Sum_V();

#if CM5
#else
  MPI_Status status;
#endif

  /*------------------ Perform segmented scan  , if any ---------------------*/

/***
if (!PLC)
print_segment(BN, no_proc, proc_list);
***/

  if ( ( BN != -1 ) && ( no_proc > 1 ) )
      COMPLEX_Segmented_Sum_V(&sendbuf2[ptr_bucket[BN]],MP_div_K,proc_list,no_proc);

  /*--------------- send/receive the preceding bucket , if any ---------------*/
  /*--------------- sending the preceeding partial bucket, if needed ---------*/

  if ( (first_bucket_partial==1) && (last_bucket_partial==1) &&
      ( first_bucket != last_bucket ))  {
    destPN = prefix_sum_count_bucket[first_bucket]/IPsize;

#if CM5

    CMMD_send_block(destPN,  CMMD_DEFAULT_TAG, 
		    &sendbuf2[ptr_bucket[first_bucket]],
		    MP_div_K*sizeof(COMPLEX_Data_Type));

#else

    MPI_Send( &sendbuf2[ptr_bucket[first_bucket]], 
	      MP_div_K*sizeof(COMPLEX_Data_Type),
	      MPI_BYTE, destPN, TAG, WORLD );

#endif

  }


  /*-------------------- receiving a bucket , if any -------------------------*/

  if ( sourcePN != -1 ) {

    temp = (COMPLEX_Data_Type *) malloc(MP_div_K*sizeof(COMPLEX_Data_Type));
    if ( temp == NULL ) {
      printf(" Out of Memory: temp in raw_stage2()\n"); 
      exit(0);
    }		

#if CM5
    CMMD_receive_block(sourcePN, CMMD_DEFAULT_TAG, 
		       temp, MP_div_K*sizeof(COMPLEX_Data_Type));

#else
    MPI_Recv( temp, MP_div_K*sizeof(COMPLEX_Data_Type), 
	      MPI_BYTE, sourcePN, TAG, WORLD, &status );

#endif

    /*for (i=0 ; i < MP_div_K ; i++)*/
      /*sendbuf2[ptr_bucket[last_bucket]+i] += temp[i];*/ /* operator */
      /*COMPLEX_comb_1(sendbuf2[ptr_bucket[last_bucket]+i], temp[i]);*/
    COMPLEX_vector_comb_1(&sendbuf2[ptr_bucket[last_bucket]], temp, MP_div_K);


    if (temp != NULL)                       free(temp);

  }

  /*-------------------     Second Stage  Communication ----------------------*/
  /*                     Combine the receive vector, recvcount,               */

  send_sendl_raw(sendcount, recvcount);

  /*--------------------   Calculating  rdispls   ----------------------------*/

  rdispls[0] = 0;
  total_recv = recvcount[0];
  for (i=1 ; i < tot_nproc ; i++)  {
    rdispls[i] = rdispls[i-1] + recvcount[i-1];
    total_recv += recvcount[i];
  }

  recvbuf2 = (COMPLEX_Data_Type *) malloc(total_recv*sizeof(COMPLEX_Data_Type));
  if ( total_recv > 0 && recvbuf2 == NULL ) {
    printf(" Out of Memory: recvbuf2 in raw_stage2()\n"); 
    exit(0);
  }		

  M_to_M_PC(sendbuf2,sendcount,sdispls,sizeof(COMPLEX_Data_Type),
	    recvbuf2,recvcount,rdispls,algorithm);

  /*-----------------   Copying the values to the final destination  ---------*/

  first = my_pnum*K;
  last_plus1 = first+K;
  j = 0;
  /*--- GENERAL ---*/
  ind = 0;
  for ( i=first ;  i < last_plus1 && ind < MP; i++)
    if ( Sparse[i] == 0 )  { /* dense bucket */
      ind = MP_div_K * (i%K );
      end = ind+MP_div_K;

      if (end > MP)
	end = MP;

      /***
      for ( ; ind < end; ind++,j++)
	target_array[ind] = recvbuf2[j];  	
      ***/

      switch (OP_TYPE) { /* vector operator */    
                        
      case COPY:            
	for ( ; ind < end; ind++,j++)
	  target_array[ind] = recvbuf2[j];  	
	break;              
                        
      case PRODUCT:         
	for ( ; ind < end; ind++,j++)
	  COMPLEX_PRODUCT(target_array[ind], recvbuf2[j]);  	
	break;              
                        
      case SUM:             
	for ( ; ind < end; ind++,j++) {
	  /***/
	  COMPLEX_SUM(target_array[ind], recvbuf2[j]);  	
	  /***/
	  /***
	  target_array[ind].Real += recvbuf2[j].Real;  	
	  target_array[ind].Imag += recvbuf2[j].Imag;  	
	  ***/
	}
	break;              
      } /* switch */

      /*for ( ; ind < end / ** && ind < MP ** /; ind++,j++)*/
	/*target_array[ind] += recvbuf2[j];*/          /* operator */
	/*COMPLEX_comb_1(target_array[ind], recvbuf2[j]);*/

    } /* if ( Sparse[i] == 0 ) */

  /*-- some case, we have segmentation fault. No idea why! --*/
  /*-- to the following at the end of receive function     --*/
  /*-- if (temp != NULL)                       free(temp); --*/

  if (recvbuf2 != NULL)                   free(recvbuf2);                     

}



/******************************************************************************
 *                                                                            *
 *  Function : COMPLEX_Segmented_Sum_V()
 *                                                                            *
 ******************************************************************************/

void COMPLEX_Segmented_Sum_V(
COMPLEX_Data_Type *vector,    /*  the initial values */
int no,               /*  the number of values, the size of vecotr is no */
                      /* the final result at proc_list[0] in vecotr */
int *proc_list,       /* the list of processors */
int no_proc)          /* the number of processors */
{
   COMPLEX_Data_Type *temp1,*temp2;
   int my_new_pnum;
   int iter,MASK;
   int partner,i,j; 
  
#if CM5
#else
  MPI_Status status;
#endif

   temp1 = (COMPLEX_Data_Type *) malloc((unsigned)(no*sizeof(COMPLEX_Data_Type)));
   temp2 = (COMPLEX_Data_Type *) malloc((unsigned)(no*sizeof(COMPLEX_Data_Type)));
   
   if ( no > 0 && temp2 == NULL )  {
     printf(" Memory Problem...in Segmented scan \n");
     exit(0);
   }
   for (i=0; i < no; i++)
     temp1[i] = vector[i];
   
   /*-- reassigning proc. ids */
   my_new_pnum = 0;
   
   while ( my_pnum != proc_list[my_new_pnum] )
     my_new_pnum++;
   
   /*iter = ceil(log(no_proc)/log(2));*/
   /*iter = (int) ceil( log2( (double)no_proc ) );*/
   iter = CLOG2(no_proc);
   
   MASK=1; /* 2**0 */
  
   for ( i=0 ; i < iter ; i++)  {
     partner = my_new_pnum ^ MASK; /* MASK = 2**i */
     MASK <<= 1;  /* left shift by 1 */
  
#if CM5

     if ( partner < no_proc )  {
     
       if ( partner < my_new_pnum )  /* send */ 
	 CMMD_send_block(proc_list[partner], CMMD_DEFAULT_TAG,
			 temp1,no*sizeof(COMPLEX_Data_Type));

       else { /* receive */  
	 CMMD_receive_block(proc_list[partner], CMMD_DEFAULT_TAG,
			    temp2,no*sizeof(COMPLEX_Data_Type));

	 /*for (j=0; j < no ; j++)*/
	   /*temp1[j] += temp2[j];*/ /* operator */
	   /*COMPLEX_comb_1(temp1[j], temp2[j]);*/
	 COMPLEX_vector_comb_1(temp1, temp2, no);
       }

     } /* if ( partner < no_proc ) */

#else

     if ( partner < no_proc )  {
       
       if ( partner < my_new_pnum )  /* send */ 
  
	 MPI_Send( temp1, no*sizeof(COMPLEX_Data_Type),
		   MPI_BYTE, proc_list[partner], TAG, WORLD );
  
       else { /* receive */  
  
	 MPI_Recv( temp2, no*sizeof(COMPLEX_Data_Type),
		   MPI_BYTE, proc_list[partner], TAG, WORLD, &status );
  
	 /*for (j=0; j < no ; j++)*/
	   /*temp1[j] += temp2[j];*/ /* operator */
	   /*COMPLEX_comb_1(temp1[j], temp2[j]);*/
	 COMPLEX_vector_comb_1(temp1, temp2, no);
       }
  
     } /* if ( partner < no_proc ) */

#endif

   } /* for */
  
   if (my_new_pnum==0)   
     for (i=0; i < no ; i++) 
       vector[i] = temp1[i];
  
   if (temp2 != NULL) free((char *) temp2);
   if (temp1 != NULL) free((char *) temp1);

}

